/*
 * $Id: Hints.java 238 2005-08-15 22:26:43Z davidb $
 * 
 * Copyright (c) 2005 VeriSign. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The name of the author may not
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package se.rfc.unbound.iter;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;

import org.apache.log4j.Logger;
import org.xbill.DNS.*;

import se.rfc.unbound.DelegationPoint;
import se.rfc.unbound.Host;
import se.rfc.unbound.SMessage;
import se.rfc.unbound.SRRset;
import se.rfc.unbound.Util;
import se.rfc.unbound.cache.CacheUtils;

/**
 * A class for managing nameserver "hints". Hints are either root hints, used
 * for finding the DNS root (on a DNS class basis), or "stub zone" hints,
 * which allow for overriding certain delegation points.
 * 
 * @author davidb
 * @version $Revision: 238 $
 */
public class Hints
{

  private static class StubHintComparator implements Comparator
  {
    public int compare(Object o1, Object o2)
    {
      Name a = (Name) o1;
      Name b = (Name) o2;
      int res = b.labels() - a.labels();
      if (res != 0) return res;

      return a.compareTo(b);
    }

  }
  private Map        mHintMap;
  private List       mStubList;
  private Comparator mStubHintComparator;

  private Logger     log = Logger.getLogger(this.getClass());

  public Hints()
  {
    mHintMap = new HashMap();
    mStubList = new ArrayList();
    mStubHintComparator = new StubHintComparator();
  }

  /**
   * Lookup a root hint for a given class. Root hints are stored as referral
   * messages so they can be returned to a client, if necessary.
   * 
   * @param dclass The class of the root hint.
   * 
   * @return a referral message for the hint, or null if the hint doesn't
   *         exist.
   */
  public SMessage lookupRootHint(int dclass)
  {
    String k = hintKey(Name.root, dclass);

    return (SMessage) mHintMap.get(k);
  }

  /**
   * Given a qname/qclass combination, and the delegation point from the cache
   * for this qname/qclass, determine if this combination indicates that a
   * stub hint exists and must be primed.
   * 
   * @param qname The qname that generated the delegation point.
   * @param qclass The qclass that generated the delegation point.
   * @param dp The cache generated delegation point.
   * @return A new priming delegation point if there is a stub hint that must
   *         be primed, otherwise null.
   */
  public DelegationPoint lookupStubHint(Name qname, int qclass,
      DelegationPoint dp)
  {
    for (Iterator i = mStubList.iterator(); i.hasNext();)
    {
      Name n = (Name) i.next();
      // First see if qname is at or underneath a stub zone.
      if (qname.subdomain(n))
      {
        // If so, fetch the stub delegation hint.
        String k = hintKey(n, DClass.IN);
        DelegationPoint stub_dp = (DelegationPoint) mHintMap.get(k);
        // If our cached delegation point is above the hint, we need to prime.
        if (stub_dp.getName().strictSubdomain(dp.getName()))
        {
          stub_dp.resetTargets();
          return stub_dp;
        }

        break;
      }
    }

    return null;
  }

  /** Calculate a key for a given hint entry. */
  private String hintKey(Name n, int dclass)
  {
    return "H" + dclass + "/" + Util.nameToString(n);
  }

  /**
   * Read in a root hints file -- this is a zone file format file with a NS
   * set for '.' and a list of A and/or AAAA records as glue. The class must
   * be expressed in the hints file -- DClass.IN will be assumed.
   * 
   * @param filename The root hints filename.
   * 
   * @throws IOException if there are problems reading the root hints file.
   */
  public void configureRootHints(String filename) throws IOException
  {
    log.debug("reading root hints file <" + filename + ">");

    SRRset ns_rrset = new SRRset();

    // First read in the whole hint file.
    Master master = new Master(filename);
    ArrayList records = new ArrayList();
    Record r = null;

    while ((r = master.nextRecord()) != null)
    {
      records.add(r);
    }

    // Then sort the records.
    // Note that Record objects sort themselves into DNSSEC canonical order.
    Collections.sort(records);

    // Now arrange them into RRsets.
    List rrsets = new LinkedList();
    rrsets.add(ns_rrset);

    SRRset cur_rrset = new SRRset();
    for (Iterator i = records.iterator(); i.hasNext();)
    {
      r = (Record) i.next();

      // Since this is a "hints" file, we assume that there is one NS rrset
      // present, the rest of the file consisting of glue.

      if (r.getType() == Type.NS)
      {
        if (r.getName().equals(Name.root))
        {
          ns_rrset.addRR(r);
        }
        // NS RRs with ownernames other than '.' will be ignored.
        continue;
      }

      // If cur_rrset is currently blank, add the current record to it.
      // This will only happen for the first non-NS rrset.
      if (cur_rrset.size() == 0)
      {
        cur_rrset.addRR(r);
        continue;
      }

      // If we are currently processing an RRset, and the current record is
      // part of it, add it.
      if (cur_rrset.getName().equals(r.getName())
          && cur_rrset.getType() == r.getType())
      {
        cur_rrset.addRR(r);
        continue;
      }

      // Otherwise, we've finished assembling the RRset, so add it to the
      // list, and reset.
      rrsets.add(cur_rrset);

      cur_rrset = new SRRset();
      cur_rrset.addRR(r);
    }
    // Make sure we add the last generated RRset.
    if (cur_rrset.size() > 0) rrsets.add(cur_rrset);

    // Convert our correctly ordered list (NS first, glue after) into a
    // referral message.
    SMessage m = CacheUtils.referralFromList(rrsets);

    String k = hintKey(Name.root, ns_rrset.getDClass());

    mHintMap.put(k, m);
  }

  /**
   * Configure stub zones. This consists of pulling stub zone hints from the
   * configuration file.
   * 
   * @param s The value of the stub zone list -- a whitespace separated list
   *          of stub zones.
   * @param prefix The property prefix of the stub zones. I.e., IP addresss
   *          for stub zone 'example.com' should be stored under property
   *          prefix.example.com.
   * @param config The configuration object. This should contain properties
   *          that get mapped to stub zone hints.
   *  
   */
  public void configureStubZones(String s, String prefix, Properties config)
  {
    // FIXME: currently stub hints are classless.

    String[] zones = s.split("\\s+");
    if (zones == null || zones.length == 0) return;

    for (int i = 0; i < zones.length; i++)
    {
      try
      {
        String z = zones[i];
        if (!z.endsWith(".")) z += ".";
        Name n = Name.fromString(z);

        String p = config.getProperty(prefix + "." + zones[i]);
        if (p == null)
        {
          log.warn("No IPs configured for stub zone " + zones[i]);
          continue;
        }

        String[] addrstrs = p.split("\\s+");
        if (addrstrs == null || addrstrs.length == 0)
        {
          log.warn("No nameservers configured for stub zone " + zones[i]);
          continue;
        }
        List targetList = new ArrayList();
        for (int j = 0; j < addrstrs.length; j++)
        {
          try
          {
            InetAddress addr = InetAddress.getByName(addrstrs[j]);
            Host h = new Host(addr);
            targetList.add(h);
          }
          catch (UnknownHostException e)
          {
            log.warn("Couldn't get address for stub zone nameserver "
                + addrstrs[j]);
            continue;
          }
        }

        log.debug("adding stub zone: " + n);
        String k = hintKey(n, DClass.IN);
        mHintMap.put(k, new DelegationPoint(n, targetList));
        mStubList.add(n);
      }
      catch (TextParseException e)
      {
        log.warn("Failed to parse stub zone name: " + zones[i], e);
      }
    }

    Collections.sort(mStubList, mStubHintComparator);
  }

}
